import unittest

import torch

from normalkl import kl

trikron = lambda A, B, C: torch.kron(A, torch.kron(B, C))

class TestKL(unittest.TestCase):
    def setUp(self):
        self.dtype = torch.float64

        d1, d2 = 3, 4
        d = d1 * d2
        tensor_d1, tensor_d2, tensor_d3 = 2, 2, 3

        self.d1 = d1
        self.d2 = d2
        self.d = d

        dtype = torch.float64

        # Full covariance
        self.L1 = torch.randn(d, d, dtype=dtype).tril()
        self.L2 = torch.randn(d, d, dtype=dtype).tril()
        self.L1_inv = torch.inverse(self.L1)
        self.L2_inv = torch.inverse(self.L2)
        self.covmat1 = self.L1 @ self.L1.T
        self.covmat2 = self.L2 @ self.L2.T

        # Cholesky factors for cholkroncov1 and cholkroncov2
        self.LU1 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV1 = torch.randn(d2, d2, dtype=dtype).tril()
        self.LU2 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV2 = torch.randn(d2, d2, dtype=dtype).tril()
        # self.LU1_inv = torch.inverse(self.LU1)
        # self.LV1_inv = torch.inverse(self.LV1)
        # self.LU2_inv = torch.inverse(self.LU2)
        # self.LV2_inv = torch.inverse(self.LV2)
        self.tensor_LU1 = torch.randn(tensor_d1, tensor_d1, dtype=dtype).tril()
        self.tensor_LV1 = torch.randn(tensor_d2, tensor_d2, dtype=dtype).tril()
        self.tensor_LW1 = torch.randn(tensor_d3, tensor_d3, dtype=dtype).tril()
        self.tensor_LU2 = torch.randn(tensor_d1, tensor_d1, dtype=dtype).tril()
        self.tensor_LV2 = torch.randn(tensor_d2, tensor_d2, dtype=dtype).tril()
        self.tensor_LW2 = torch.randn(tensor_d3, tensor_d3, dtype=dtype).tril()

        # Covariance matrices for cholkroncov1 and cholkroncov2
        self.U1 = self.LU1 @ self.LU1.T
        self.V1 = self.LV1 @ self.LV1.T
        self.U2 = self.LU2 @ self.LU2.T
        self.V2 = self.LV2 @ self.LV2.T
        self.U1_inv = torch.inverse(self.U1)
        self.V1_inv = torch.inverse(self.V1)
        self.U2_inv = torch.inverse(self.U2)
        self.V2_inv = torch.inverse(self.V2)

        self.tensor_U1 = self.tensor_LU1 @ self.tensor_LU1.T
        self.tensor_V1 = self.tensor_LV1 @ self.tensor_LV1.T
        self.tensor_W1 = self.tensor_LW1 @ self.tensor_LW1.T
        self.tensor_U2 = self.tensor_LU2 @ self.tensor_LU2.T
        self.tensor_V2 = self.tensor_LV2 @ self.tensor_LV2.T
        self.tensor_W2 = self.tensor_LW2 @ self.tensor_LW2.T

        # Cholesky Kronecker decompositions
        self.cholkroncov1 = (self.LU1, self.LV1)
        self.cholkroncov2 = (self.LU2, self.LV2)
        self.trikron_covmat1 = trikron(self.tensor_U1, self.tensor_V1, self.tensor_W1)
        self.trikron_covmat2 = trikron(self.tensor_U2, self.tensor_V2, self.tensor_W2)

        # Full covariance matrices
        self.kron_covmat1 = torch.kron(self.U1, self.V1)
        self.kron_covmat2 = torch.kron(self.U2, self.V2)
        self.kron_precmat1 = torch.kron(self.U1_inv, self.V1_inv)
        self.kron_precmat2 = torch.kron(self.U2_inv, self.V2_inv)

        # Other inputs
        self.I = torch.eye(d, dtype=dtype)
        self.v1 = torch.randn(d, dtype=dtype)
        self.v2 = torch.randn(d, dtype=dtype)
        self.s2 = torch.tensor(2.0, dtype=dtype)
        self.a2 = torch.randn(d1, dtype=dtype).square()
        self.b2 = torch.randn(d2, dtype=dtype).square()

        # Diagonal covariance and precision
        self.diagvar = torch.diag(self.covmat1)
        self.diagprec = 1.0 / self.diagvar

        # Scalar variance and precision
        self.scalarvar = torch.mean(self.diagvar).unsqueeze(0)
        self.scalarprec = 1.0 / self.scalarvar

        # Precision matrix
        self.precmat1 = torch.inverse(self.covmat1)
        self.precmat2 = torch.inverse(self.covmat2)

        self.covs2 = [('covmat', self.covmat2, self.covmat2),
                      ('precmat', self.precmat2, self.covmat2),
                      ('diagvar', self.diagvar, self.diagvar.diag()),
                      ('diagprec', self.diagprec, self.diagvar.diag()),
                      ('scalarvar', self.scalarvar, self.scalarvar * self.I),
                      ('scalarprec', self.scalarprec, self.scalarvar * self.I),
                      ('identity', None, self.I),
                      ('cholcov', self.L2, self.covmat2),
                      ('cholprec', torch.cholesky(self.precmat2), self.covmat2),
                      ('kroncov', (self.U2, self.V2), self.kron_covmat2),
                      ('kronprec', (self.U2_inv, self.V2_inv), self.kron_covmat2),
                      ('cholkroncov', (self.LU2, self.LV2), self.kron_covmat2),
                      ('choltrikroncov', (self.tensor_LU2, self.tensor_LV2, self.tensor_LW2), self.trikron_covmat2),
                      # ('diagvarkron', (self.a2, self.b2), torch.kron(self.a2.diag(), self.b2.diag())),
                      # ('diagvarrow', self.a2, torch.kron(torch.eye(self.d1), self.b2.diag())),
                      # ('diagvarcov', self.b2, torch.kron(self.a2.diag(), torch.eye(self.d2)))
                      ]

        # Tolerances
        self.rtol = 1e-6
        self.atol = 1e-6

    def test_kl_readme_example(self):
        dtype = self.dtype
        mean1 = torch.tensor([4.0, 5.0], dtype=dtype)
        covariance1 = torch.tensor([[1.0, 1.0], [2.0, 4.0]], dtype=dtype)
        mean2 = torch.tensor([1.0, 2.0], dtype=dtype)
        scalarvar2 = torch.tensor([3.0], dtype=dtype)

        result = kl(mean1, 'covmat', covariance1, mean2, 'scalarvar', scalarvar2)

        expected = torch.tensor([3.5853720317214703], dtype=dtype)
        self.assertTrue(torch.allclose(expected, result))

    def test_kl_covmat_any(self):
        for cov_type2, cov2, expected_covmat2 in self.covs2:
            with self.subTest(case=cov_type2):
                expected = kl(self.v1, 'covmat', self.covmat1, self.v2, 'covmat', expected_covmat2)
                result = kl(self.v1, 'covmat', self.covmat1, self.v2, cov_type2, cov2)

                self.assertTrue(torch.allclose(expected, result), f"{cov_type2} fails: {expected:.4f}!={result:.4f}")



# Run the tests
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




import unittest

import torch

from normalkl import kl, optimal_covariance

trikron = lambda A, B, C: torch.kron(A, torch.kron(B, C))

class TestOptimalCov2(unittest.TestCase):
    def setUp(self):
        torch.manual_seed(0)

        dtype = torch.float64

        d1, d2 = 3, 4
        d = d1 * d2
        self.d1 = d1
        self.d2 = d2
        self.d = d

        tensor_d1, tensor_d2, tensor_d3 = 2, 2, 3
        self.tensor_d1 = tensor_d1
        self.tensor_d2 = tensor_d2
        self.tensor_d3 = tensor_d3

        dtype=torch.float64

        # Means
        self.mean1 = torch.randn(d, dtype=dtype)
        self.mean2 = torch.randn(d, dtype=dtype)

        # Full covariance 
        self.L1 = torch.randn(d, d, dtype=dtype).tril()
        self.L2 = torch.randn(d, d, dtype=dtype).tril()
        self.covmat1 = self.L1 @ self.L1.T
        self.covmat2 = self.L2 @ self.L2.T
            
        # Cholesky factors for cholkroncov1 and cholkroncov2
        self.LU1 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV1 = torch.randn(d2, d2, dtype=dtype).tril()
        self.LU2 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV2 = torch.randn(d2, d2, dtype=dtype).tril()

        self.tensor_LU1 = torch.randn(tensor_d1, tensor_d1, dtype=dtype).tril()
        self.tensor_LV1 = torch.randn(tensor_d2, tensor_d2, dtype=dtype).tril()
        self.tensor_LW1 = torch.randn(tensor_d3, tensor_d3, dtype=dtype).tril()
        self.tensor_LU2 = torch.randn(tensor_d1, tensor_d1, dtype=dtype).tril()
        self.tensor_LV2 = torch.randn(tensor_d2, tensor_d2, dtype=dtype).tril()
        self.tensor_LW2 = torch.randn(tensor_d3, tensor_d3, dtype=dtype).tril()

        # Covariance matrices for cholkroncov1 and cholkroncov2
        self.U1 = self.LU1 @ self.LU1.T
        self.V1 = self.LV1 @ self.LV1.T
        self.U2 = self.LU2 @ self.LU2.T
        self.V2 = self.LV2 @ self.LV2.T

        self.tensor_U1 = self.tensor_LU1 @ self.tensor_LU1.T
        self.tensor_V1 = self.tensor_LV1 @ self.tensor_LV1.T
        self.tensor_W1 = self.tensor_LW1 @ self.tensor_LW1.T
        self.tensor_U2 = self.tensor_LU2 @ self.tensor_LU2.T
        self.tensor_V2 = self.tensor_LV2 @ self.tensor_LV2.T
        self.tensor_W2 = self.tensor_LW2 @ self.tensor_LW2.T

        # Cholesky Kronecker decompositions
        self.cholkroncov1 = (self.LU1, self.LV1)
        self.cholkroncov2 = (self.LU2, self.LV2)

        self.choltrikroncov1 = (self.tensor_LU1, self.tensor_LV1, self.tensor_LW1)
        self.choltrikroncov2 = (self.tensor_LU2, self.tensor_LV2, self.tensor_LW1)

        # Full covariance matrices
        self.kron_covmat1 = torch.kron(self.U1, self.V1)
        self.kron_covmat2 = torch.kron(self.U2, self.V2)
        self.trikron_covmat1 = trikron(self.tensor_U1, self.tensor_V1, self.tensor_W1)
        self.trikron_covmat2 = trikron(self.tensor_U2, self.tensor_V2, self.tensor_W2)

        # Other inputs
        self.I = torch.eye(d, dtype=dtype)
        self.v1 = torch.randn(d, dtype=dtype)
        self.v2 = torch.randn(d, dtype=dtype)
        self.s2 = torch.tensor(2.0, dtype=dtype)
        self.a1 = torch.randn(d1, dtype=dtype).square()
        self.b1 = torch.randn(d2, dtype=dtype).square()

        self.rtol = 1e-6
        self.atol = 1e-6

        self.cov1_list = [('covmat', self.covmat1),
                     ('precmat', self.covmat1),
                     ('diagvar', self.v1),
                     ('diagprec', self.v1),
                     ('scalarvar', self.s2),
                     ('scalarprec', self.s2),
                     ('identity', None),
                     ('cholcov', self.L1),
                     ('cholprec', self.L1),
                     ('kroncov', (self.U1, self.V1)),
                     ('kronprec', (self.U1, self.U2)),
                     ('cholkroncov', (self.LU1, self.LV1)),
                     ('cholkronprec', (self.LU1, self.LV1)),
                     ('diagvarkron', (self.a1, self.b1)),
                     ('diagvarrow', self.a1),
                     ('diagvarcol', self.b1)]



    def _is_optimum(self, objective_function, point, tolerance=1e-3):
        """
        Check if the point is at an optimum of the objective function.

        Args:
            point (torch.Tensor): The point to check.
            objective_function (callable): The objective function to minimize or maximize.
            tolerance (float): The tolerance for checking if the gradient is close to zero.

        Returns:
            bool: True if the point is at the optimum, False otherwise.
        """
        # Ensure the point requires gradients for optimization checking
        point = point.clone().detach().requires_grad_(True)

        # Compute the objective function value at the given point
        obj_value = objective_function(point)
        
        # Compute the gradients with respect to the point
        obj_value.backward()

        # Check if the gradients are close to zero within the specified tolerance
        is_at_optimum = torch.all(torch.abs(point.grad) < tolerance)

        return is_at_optimum

    # test_optimal_*_covmat

    def test_optimal_covmat_covmat(self):
        optimal_covmat2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'covmat')
        kl_func = lambda covmat2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'covmat', covmat2)

        self.assertTrue(self._is_optimum(kl_func, optimal_covmat2))

    def test_optimal_cholkroncov_covmat(self):
        optimal_covmat2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'covmat')
        kl_func = lambda covmat2: kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'covmat', covmat2)

        self.assertTrue(self._is_optimum(kl_func, optimal_covmat2))

    # test_optimal_*_precmat

    def test_optimal_covmat_precmat(self):
        optimal_precmat2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'precmat')
        kl_func = lambda precmat2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'precmat', precmat2)

        self.assertTrue(self._is_optimum(kl_func, optimal_precmat2))

    def test_optimal_cholkroncov_precmat(self):
        optimal_precmat2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'precmat')
        kl_func = lambda precmat2: kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'precmat', precmat2)

        self.assertTrue(self._is_optimum(kl_func, optimal_precmat2))

    # test_optimal_*_diagprec

    def test_optimal_covmat_diagprec(self):
        optimal_diagprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec')
        kl_func = lambda diagprec2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec', diagprec2)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagprec2))


    # test_optimal_*_diagvar

    def test_optimal_covmat_diagvar(self):
        optimal_diagvar2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar')
        kl_func = lambda diagvar2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar', diagvar2)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagvar2))

    # test_optimal_*_diagprec

    def test_optimal_covmat_diagprec(self):
        optimal_diagprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec')
        kl_func = lambda diagprec2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec', diagprec2)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagprec2))

    # test_optimal_*_scalarvar

    def test_optimal_covmat_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar')
        kl_func = lambda scalarvar2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar', scalarvar2)

        self.assertTrue(self._is_optimum(kl_func, optimal_scalarvar2))

    def test_optimal_cholkroncov_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar')
        kl_func = lambda scalarvar2: kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar', scalarvar2)

        self.assertTrue(self._is_optimum(kl_func, optimal_scalarvar2))

    # test_optimal_*_scalarprec

    def test_optimal_covmat_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec')
        kl_func = lambda scalarprec2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec', scalarprec2)

        self.assertTrue(self._is_optimum(kl_func, optimal_scalarprec2))


    def test_optimal_cholkroncov_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec')
        kl_func = lambda scalarprec2: kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec', scalarprec2)

        self.assertTrue(self._is_optimum(kl_func, optimal_scalarprec2))

    # test_optimal_*_diagvarrow

    def test_optimal_covmat_diagvarrow(self):
        optimal_diagvarrow2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarrow', dims=(self.d1, self.d2))
        kl_func = lambda diagvarrow2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarrow', diagvarrow2)

        #self.assertTrue(False)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagvarrow2))

    def test_optimal_cholkroncov_diagvarrow(self):
        optimal_diagvarrow2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvarrow', dims=(self.d1, self.d2))
        kl_func = lambda diagvarrow2: kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvarrow', diagvarrow2)

        #self.assertTrue(False)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagvarrow2))

    def test_optimal_choltrikroncov_diagvarrow(self):
        optimal_diagvarrow2 = optimal_covariance(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'diagvarrow', dims=(self.tensor_d1, self.tensor_d2, self.tensor_d3))
        kl_func = lambda diagvarrow2: kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'diagvarrow', diagvarrow2)

        #self.assertTrue(False)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagvarrow2))


    # test_optimal_*_diagvarrow


    def test_optimal_covmat_diagvarcol(self):
        optimal_diagvarcol2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarcol', dims=(self.d1, self.d2))
        kl_func = lambda diagvarcol2: kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarcol', diagvarcol2)

        #self.assertTrue(False)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagvarcol2))

    def test_optimal_cholkroncov_diagvarcol(self):
        optimal_diagvarcol2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvarcol', dims=(self.d1, self.d2))
        kl_func = lambda diagvarcol2: kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvarcol', diagvarcol2)

        #self.assertTrue(False)

        self.assertTrue(self._is_optimum(kl_func, optimal_diagvarcol2))



# Run the tests
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




import torch
import math

from .logdet import *

from normalkl.util import dim_dtype_device_of_cov, is_zero

trikron = lambda A, B, C: torch.kron(A, torch.kron(B, C))

# auto_kl_*_covmat
def auto_kl_covmat_covmat(mean1, covmat1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    mean_diff = mean1 - mean2
    return 0.5 * torch.log(1 + (mean_diff.unsqueeze(0) @ torch.inverse(covmat1) @ mean_diff)).squeeze()

def auto_kl_precmat_covmat(mean1, precmat1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    mean_diff = mean1 - mean2
    return 0.5 * torch.log(1 + (mean_diff.unsqueeze(0) @ precmat1 @ mean_diff).squeeze())

def auto_kl_cholkroncov_covmat(mean1, cholkroncov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T)) || M(v2, diag(v3)) )
    """
    mean_diff = mean1 - mean2
    LU1, LV1 = cholkroncov1
    d1, d2 = LU1.size(0), LV1.size(0)

    M = mean_diff.view(d1, d2)
    
    # Solve LU1^{-1} * M * LV1^{-T} using triangular solves
    W_A = torch.triangular_solve(M, LU1, upper=False)[0]  # LU1^{-1} * M
    W = torch.triangular_solve(W_A.t(), LV1, upper=False)[0]  # LV1^{-1} * W_A^T
    W = W.t()  # Transpose back to the correct shape
    
    # Compute the Frobenius norm of W (which is equivalent to the quadratic form)
    result = torch.sum(W ** 2)

    return 0.5 * torch.log(1 + result)

# auto_kl_*_precmat
def auto_kl_covmat_precmat(mean1, covmat1, mean2):
    return auto_kl_covmat_covmat(mean1, covmat1, mean2)

def auto_kl_precmat_precmat(mean1, precmat1, mean2):
    return auto_kl_precmat_covmat(mean1, precmat1, mean2)

def auto_kl_cholkroncov_precmat(mean1, cholkroncov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T)) || M(v2, M2))
    """
    return auto_kl_cholkroncov_covmat(mean1, cholkroncov1, mean2)

# auto_kl_*_diagvar
def auto_kl_covmat_diagvar(mean1, covmat1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    mean_diff_square = (mean1 - mean2).square()

    covmat1_diag = covmat1.diag()
    return 0.5 * (torch.sum(torch.log(covmat1_diag + mean_diff_square)) - logdet_covmat(covmat1))

def auto_kl_cholkroncov_diagvar(mean1, cholkroncov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    mean_diff_square = (mean1 - mean2).square()

    LU1, LV1 = cholkroncov1
    covmat1_diag = torch.kron(LU1.square().sum(1), LV1.square().sum(1))

    return 0.5 * (torch.sum(torch.log(covmat1_diag + mean_diff_square)) - logdet_cholkroncov(cholkroncov1))

def auto_kl_choltrikroncov_diagvar(mean1, choltrikroncov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    mean_diff_square = (mean1 - mean2).square()

    LU1, LV1, LW1 = choltrikroncov1
    covmat1_diag = trikron(LU1.square().sum(1), LV1.square().sum(1), LW1.square().sum(1))

    return 0.5 * (torch.sum(torch.log(covmat1_diag + mean_diff_square)) - logdet_choltrikroncov(choltrikroncov1))

# auto_kl_*_diagprec
def auto_kl_covmat_diagprec(mean1, covmat1, mean2):
    return auto_kl_covmat_diagvar(mean1, covmat1, mean2)

# auto_kl_*_scalarvar
def auto_kl_covmat_scalarvar(mean1, covmat1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, v3 I))
    """
    d = mean1.size(0)
    mean_diff_square = (mean1 - mean2).square()

    total_variance = torch.trace(covmat1) + mean_diff_square.sum()
    return 0.5 * (d * torch.log(total_variance / d) - logdet_covmat(covmat1))

def auto_kl_cholcov_scalarvar(mean1, cholcov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, L1 L1.T) || M(v2, v3 I))
    """
    d = mean1.size(0)
    mean_diff_square = (mean1 - mean2).square()

    covmat1_trace = torch.sum(cholcov1 ** 2)

    total_variance = covmat1_trace + mean_diff_square.sum()
    return 0.5 * (d * torch.log(total_variance / d) - logdet_cholcov(cholcov1))

def auto_kl_cholkroncov_scalarvar(mean1, cholkroncov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, v3 I))
    """
    d = mean1.size(0)
    LU, LV = cholkroncov1
    cov1_trace = LU.square().sum() * LV.square().sum()

    mean_diff_square = (mean1 - mean2).square()

    total_variance = cov1_trace + mean_diff_square.sum()
    return 0.5 * (d * torch.log(total_variance / d) - logdet_cholkroncov(cholkroncov1))

def auto_kl_choltrikroncov_scalarvar(mean1, choltrikroncov1, mean2):
    """ Compute KL with automatic optimal diagonal variance according to
        argmin_v3 KL( N(v1, M1) || M(v2, v3 I))
    """
    d = mean1.size(0)
    LU, LV, LW = choltrikroncov1
    cov1_trace = LU.square().sum() * LV.square().sum() * LW.square().sum()

    mean_diff_square = (mean1 - mean2).square()

    total_variance = cov1_trace + mean_diff_square.sum()
    return 0.5 * (d * torch.log(total_variance / d) - logdet_choltrikroncov(choltrikroncov1))

# auto_kl_*_scalarprec
def auto_kl_covmat_scalarprec(mean1, covmat1, mean2):
    return auto_kl_covmat_scalarvar(mean1, covmat1, mean2)

def auto_kl_cholkroncov_scalarprec(mean1, cholkroncov1, mean2):
    return auto_kl_cholkroncov_scalarvar(mean1, cholkroncov1, mean2)

def auto_kl_choltrikroncov_scalarprec(mean1, choltrikroncov1, mean2):
    return auto_kl_choltrikroncov_scalarvar(mean1, choltrikroncov1, mean2)

# auto_kl_*_diagvarrow
def auto_kl_covmat_diagvarrow(mean1, covmat1, mean2, dims):
    d1, d2 = dims
    both_means_float = is_zero(mean1) and is_zero(mean2)
    mean_diff_square = (mean1 - mean2).square()
    mean_diff_rows = mean_diff_square.view(d1, d2).sum(1)
    covmat1_rows = torch.diagonal(covmat1.view(d1, d2, d1, d2), dim1=1, dim2=3).sum(-1)

    optimal_diagvarrow = (covmat1_rows.diag() + mean_diff_rows) / d2

    return 0.5 * (d2 * torch.log(optimal_diagvarrow).sum() - logdet_covmat(covmat1))

def auto_kl_cholkroncov_diagvarrow(mean1, cholkroncov1, mean2, dims):
    d1, d2 = dims
    both_means_float = is_zero(mean1) and is_zero(mean2)
    mean_diff_square = (mean1 - mean2).square()
    mean_diff_rows = mean_diff_square.view(d1, d2).sum(1)

    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T

    covmat1_rows = U1 * LV1.square().sum()

    optimal_diagvarrow = (covmat1_rows.diag() + mean_diff_rows) / d2

    return 0.5 * (d2 * torch.log(optimal_diagvarrow).sum() - logdet_cholkroncov(cholkroncov1))

def auto_kl_choltrikroncov_diagvarrow(mean1, choltrikroncov1, mean2, dims):
    d1, d2, d3 = dims
    both_means_float = is_zero(mean1) and is_zero(mean2)
    mean_diff_square = (mean1 - mean2).square()
    mean_diff_rows = mean_diff_square.view(d1, d2, d3).sum((1, 2))

    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T

    covmat1_rows = U1 * LV1.square().sum() * LW1.square().sum()

    print(d1, d2, d3, covmat1_rows.shape, mean_diff_rows.shape)
    optimal_diagvarrow = (covmat1_rows.diag() + mean_diff_rows) / d2 / d3

    return 0.5 * (d2 * d3 * torch.log(optimal_diagvarrow).sum() - logdet_choltrikroncov(choltrikroncov1))

# auto_kl_*_diagvarcol
def auto_kl_covmat_diagvarcol(mean1, covmat1, mean2, dims):
    d1, d2 = dims
    both_means_float = is_zero(mean1) and is_zero(mean2)
    mean_diff_square = (mean1 - mean2).square()
    mean_diff_cols = mean_diff_square.view(d1, d2).sum(0)
    covmat1_cols = torch.diagonal(covmat1.view(d1, d2, d1, d2), dim1=0, dim2=2).sum(-1)

    optimal_diagvarrow = (covmat1_cols.diag() + mean_diff_cols) / d1

    return 0.5 * (d1 * torch.log(optimal_diagvarcol).sum() - logdet_covmat(covmat1))

def auto_kl_cholkroncov_diagvarcol(mean1, cholkroncov1, mean2, dims):
    d1, d2 = dims
    mean_diff_square = (mean1 - mean2).square()
    mean_diff_cols = mean_diff_square.view(d1, d2).sum(0)

    LU1, LV1 = cholkroncov1
    V1 = LV1 @ LV1.T

    covmat1_cols = V1 * LU1.square().sum()

    optimal_diagvarcol = (covmat1_cols.diag() + mean_diff_cols) / d1

    return 0.5 * (d1 * torch.log(optimal_diagvarcol).sum() - logdet_cholkroncov(cholkroncov1))


def auto_kl(mean1, cov_type1, cov1, mean2, cov_type2, dims=None):
    """ Compute KL with automatic optimal covariance according to
            argmin_cov2 KL( N(mean1, cov1) || N(mean2, cov2))
        and constrained by type of covariance (cov_type2)
    """
    func_name = f"auto_kl_{cov_type1}_{cov_type2}"

    func = globals()[func_name]

    total_dim, dtype, device = dim_dtype_device_of_cov(mean1, cov_type1, cov1)

    if isinstance(mean1, float):
        mean1 = torch.zeros(total_dim, dtype=dtype, device=device)
    if isinstance(mean2, float):
        mean2 = torch.zeros(total_dim, dtype=dtype, device=device)

    if cov_type2 in ['diagvarkron', 'diagvarrow', 'diagvarcol']:
        return func(mean1, cov1, mean2, dims)
    else:
        return func(mean1, cov1, mean2)





""" Functions that compute quadratic term of the KL divergence for different covariance types. """

import torch



def is_zero(x):
    if isinstance(x, float):
        return x == 0.0
    elif isinstance(x, torch.Tensor):
        return torch.equal(x, torch.tensor(0.0)) and x.numel() == 1
    else:
        return False


# Precision versions of the above functions:

def quadratic_covmat(v1, covmat2):
    """ Return quadratic term: v1^T covmat22^{-1} v1 / 2 """

    return 0.5 * (v1.unsqueeze(0) @ torch.inverse(covmat2) @ v1).squeeze()

def quadratic_precmat(v1, precmat2):
    """ Return quadratic term: v1^T precmat2 v1 / 2 """
    return 0.5 * (v1.unsqueeze(0) @ precmat2 @ v1).squeeze()

def quadratic_diagvar(v1, diagvar2):
    """ Return quadratic term: v1^T diag(var)^{-1} v1 / 2"""
    v1_square = 0.0 if is_zero(v1) else v1.square()
    return 0.5 * torch.sum(v1_square / diagvar2).squeeze()

def quadratic_diagprec(v1, diagprec2):
    """ Return quadratic term: v1^T diag(prec2) v1 / 2 """
    v1_square = 0.0 if is_zero(v1) else v1.square()
    return 0.5 * torch.sum(v1_square * diagprec2).squeeze()

def quadratic_scalarvar(v1, scalarvar2):
    """ Return quadratic term: v1^T (scalarvar2 * I)^{-1} v1 / 2 """
    v1_square = 0.0 if is_zero(v1) else v1.square()
    return 0.5 * torch.sum(v1_square / scalarvar2).squeeze()

def quadratic_scalarprec(v1, scalarprec2):
    """ Return quadratic term: v1^T (scalarprec * I) v1 / 2 """
    v1_square = 0.0 if is_zero(v1) else v1.square()
    return 0.5 * torch.sum(v1_square * scalarprec2).squeeze()

def quadratic_identity(v1, _ignore):
    """ Return quadratic term: v1^T v1 / 2 """
    v1_square = 0.0 if is_zero(v1) else v1.square()
    return 0.5 * v1_square.sum().squeeze()

def quadratic_cholcov(v1, cholcov2):
    """ Return quadratic term: v1^T (L2 L2.T )^{-1} v1 / 2 
    = v1^T L2^{-1}^T L2^{-1} v1 / 2 
    = (L2^{-1} v1)^2 / 2
    """
    L_inv = torch.inverse(cholcov2) 
    return 0.5 * torch.sum((L_inv @ v1) ** 2)

def quadratic_cholprec(v1, cholprec2):
    """ Return quadratic term: v1^T (L2 L2^T) v1 / 2 
    = v1^T L2 (L2^T v1)
    = (L2^T v1)^T (L2^T v1)
    """
    LTv = cholprec2.T @ v1
    return 0.5 * torch.sum(LTv.square())

def quadratic_kroncov(v1, kroncov2):
    """ Return quadratic term: v1^T (U2 kron V2)^{-1} v1 / 2 """
    U2, V2 = kroncov2
    d1, d2 = len(U2), len(V2)
    v1_mat = v1.view(d1, d2)  # Reshape v1 to a matrix of shape (d1, d2)
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    return 0.5 * torch.trace(U2_inv @ v1_mat @ V2_inv @ v1_mat.T)

def quadratic_kronprec(v1, kronprec2):
    """ Return quadratic term: v1^T (U2 kron V2) v1 / 2 """
    U2, V2 = kronprec2
    d1, d2 = len(U2), len(V2)
    v1_mat = v1.view(d1, d2)  # Reshape v1 to a matrix of shape (d1, d2)
    return 0.5 * torch.trace(U2 @ v1_mat @ V2 @ v1_mat.T)

def quadratic_cholkroncov(v1, cholkroncov2):
    """ Return quadratic term: v1^T (L_U2 L_U2.T kron L_V2 L_V2.T)^{-1} v1 / 2 """
    LU2, LV2 = cholkroncov2
    d1, d2 = len(LU2), len(LV2)
    v1_mat = v1.view(d1, d2)  # Reshape v1 to a matrix of shape (d1, d2)
    LU2_inv = torch.inverse(LU2)
    LV2_inv = torch.inverse(LV2)
    return 0.5 * v1.unsqueeze(0) @ torch.kron(LU2_inv.T @ LU2_inv, LV2_inv.T @ LV2_inv) @ v1

def quadratic_choltrikroncov(v1, choltrikroncov2):
    """ Return quadratic term: v1^T (L_U2 L_U2.T kron L_V2 L_V2.T kron L_W2 LW2.T)^{-1} v1 / 2 """
    LU2, LV2, LW2 = choltrikroncov2
    d1, d2, d3 = len(LU2), len(LV2), len(LW2)
    v1_tensor = v1.view(d1, d2, d3)  # Reshape v1 to a tensor of shape (d1, d2, d3)
    LU2_inv = torch.inverse(LU2)
    LV2_inv = torch.inverse(LV2)
    LW2_inv = torch.inverse(LW2)
    trikron = lambda A, B, C: torch.kron(A, torch.kron(B, C))
    return 0.5 * v1.unsqueeze(0) @ trikron(LU2_inv.T @ LU2_inv, LV2_inv.T @ LV2_inv, LW2_inv.T @ LW2_inv) @ v1

def quadratic_diagvarkron(v1, a2, b2):
    """ Return quadratic term: v1^T diag(vec(b2 a2^T))^{-1} v1 / 2 """
    v1_square = 0.0 if is_zero(v1) else v1.square()
    ab = torch.outer(a2, b2)
    ab_diag = ab.view(-1)
    return 0.5 * torch.sum(v1_square / ab_diag)

def quadratic_diagvarrow(v1, a2):
    """ Return quadratic term: v1^T diag(vec(1 a2^T))^{-1} v1 / 2 """
    v1_square = 0.0 if is_zero(v1) else v1.square()
    total_dim = len(v1)
    d1 = len(a2)
    d2 = total_dim // d1

    return 0.5 * torch.sum(v1_square.view(d1, d2) / a2.view(-1, 1))

def quadratic_diagvarcol(v1, b2):
    """ Return quadratic term: v1^T diag(vec(b2 1^T))^{-1} v1 / 2"""
    v1_square = 0.0 if is_zero(v1) else v1.square()

    total_dim = len(v1)
    d2 = len(b2)
    d1 = total_dim // d2
    return 0.5 * torch.sum(v1_square.view(d1, d2) / b2.view(1, -1))








""" Functions that compute log-determinant term of the KL divergence for different covariance types. """

import torch
import math

def logdet_covmat(covmat):
    """ Log determinant for full covariance matrix """
    return torch.logdet(covmat)

def logdet_precmat(precmat):
    """ Log determinant for full precision matrix """
    return -torch.logdet(precmat)

def logdet_diagvar(diagvar):
    """ Log determinant for diagonal covariance matrix (vector) """
    return diagvar.abs().log().sum().squeeze()

def logdet_scalarvar(scalarvar, total_dim):
    """ Log determinant for scalar variance """
    return total_dim * scalarvar.abs().log().squeeze()

def logdet_cholcov(cholcov):
    """ Log determinant for Cholesky covariance matrix """
    L = cholcov
    return L.diag().square().log().sum()

def logdet_identity(_ignore):
    """ Log determinant of identity is zero since det(I)=1 for any dimensionality """
    return 0

def logdet_kroncov(kroncov):
    """ Log determinant for Kronecker product of two matrices """
    U, V = kroncov
    return len(V) * logdet_covmat(U) + len(U) * logdet_covmat(V)

def logdet_cholkroncov(cholkroncov):
    LU, LV = cholkroncov
    """ Log determinant for Cholesky Kronecker covariance """
    return len(LV) * logdet_cholcov(LU) + len(LU) * logdet_cholcov(LV)

def logdet_choltrikroncov(choltrikroncov):
    LU, LV, LW = choltrikroncov
    """ Log determinant for Cholesky Kronecker covariance """
    return len(LV) * len(LW) * logdet_cholcov(LU) + len(LU) * len(LW) * logdet_cholcov(LV) + len(LU) * len(LV) * logdet_cholcov(LW)

def logdet_diagvarkron(diagvarkron):
    """ Log determinant for diagonal row+col (diag(u) kron diag(v)) diagonal matrix """
    rowvar, colvar = diagvarkron
    d1, d2 = rowvar.size(0), colvar.size(0)

    return d2 * torch.log(rowvar).sum() + d1 * torch.log(colvar).sum()

def logdet_diagvarrow(diagvarrow, total_dim):
    """ Log determinant for diagonal row diagonal matrix """
    row_dim = diagvarrow.size(0)
    nonrow_dim = total_dim // row_dim

    return nonrow_dim * torch.log(diagvarrow).sum()

def logdet_diagvarcol(diagvarcol, total_dim):
    """ Log determinant for diagonal col diagonal matrix """
    col_dim = diagvarcol.size(0)
    noncol_dim = total_dim // col_dim

    return noncol_dim * torch.log(diagvarcol).sum()


# Precision versions of the above functions:

def logdet_precmat(precmat):
    """ Log determinant for precision matrix """
    return -logdet_covmat(precmat)

def logdet_diagprec(diagprec):
    """ Log determinant for diagonal precision matrix """
    return -diagprec.abs().log().sum()

def logdet_scalarprec(scalarprec, total_dim):
    """ Log determinant for scalar precision """
    return -total_dim * scalarprec.log()

def logdet_cholprec(cholprec):
    """ Log determinant for Cholesky precision matrix """
    L = cholprec
    return -L.diag().square().log().sum()

def logdet_kronprec(kronprec):
    U, V = kronprec
    """ Log determinant for Kronecker product of two precision matrices """
    return len(V) * logdet_precmat(U) + len(U) * logdet_precmat(V)




import torch

trikron = lambda A, B, C: torch.kron(A, torch.kron(B, C))

def is_zero(x):
    if isinstance(x, float):
        return x == 0.0
    elif isinstance(x, torch.Tensor):
        return torch.equal(x, torch.tensor(0.0)) and x.numel() == 1
    else:
        return False

# optimal_*_covmat
def optimal_covmat_covmat(mean_diff, covmat1):
    """ Compute optimal diagonal variance of
        argmin_M2 KL( N(v1, M1) || M(v2, M2))
    """
    mean_diff_outer = 0.0 if is_zero(mean_diff) else torch.outer(mean_diff, mean_diff) 

    return covmat1 + mean_diff_outer

def optimal_precmat_covmat(mean_diff, precmat1):
    """ Compute optimal diagonal variance of
        argmin_M2 KL( N(v1, M1) || M(v2, M2))
    """
    mean_diff_outer = 0.0 if is_zero(mean_diff) else torch.outer(mean_diff, mean_diff) 

    return torch.inverse(precmat1) + mean_diff_outer

def optimal_cholkroncov_covmat(mean_diff, cholkroncov1):
    """ Compute optimal diagonal variance of
        argmin_M2 KL( N(v1, M1) || M(v2, M2))
    """
    mean_diff_outer = 0.0 if is_zero(mean_diff) else torch.outer(mean_diff, mean_diff) 

    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    covmat1 = torch.kron(U1, V1)
    return covmat1 + mean_diff_outer

def optimal_choltrikroncov_covmat(mean_diff, choltrikroncov1):
    """ Compute optimal diagonal variance of
        argmin_M2 KL( N(v1, M1) || M(v2, M2))
    """
    mean_diff_outer = 0.0 if is_zero(mean_diff) else torch.outer(mean_diff, mean_diff) 

    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    W1 = LW1 @ LW1.T
    covmat1 = trikron(U1, V1, W1)
    return covmat1 + mean_diff_outer


# optimal_*_precmat
def optimal_covmat_precmat(mean_diff, covmat1):
    """ Compute optimal diagonal variance of
        argmin_M2 KL( N(v1, M1) || M(v2, M2))
    """
    mean_diff_outer = 0.0 if is_zero(mean_diff) else torch.outer(mean_diff, mean_diff) 

    return torch.inverse(covmat1 + mean_diff_outer)

def optimal_precmat_precmat(mean_diff, covmat1):
    """ Compute optimal diagonal variance of
        argmin_M2 KL( N(v1, M1) || M(v2, M2))
    """
    mean_diff_outer = 0.0 if is_zero(mean_diff) else torch.outer(mean_diff, mean_diff) 

    return torch.inverse(torch.inverse(covmat1) + mean_diff_outer)

def optimal_cholkroncov_precmat(mean_diff, cholkroncov1):
    """ Compute optimal diagonal variance of
        argmin_M2 KL( N(v1, M1) || M(v2, M2))
    """
    mean_diff_outer = 0.0 if is_zero(mean_diff) else torch.outer(mean_diff, mean_diff) 
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    covmat1 = torch.kron(U1, V1)
    return torch.inverse(covmat1 + mean_diff_outer)

def optimal_choltrikroncov_precmat(mean_diff, choltrikroncov1):
    """ Compute optimal diagonal variance of
        argmin_M2 KL( N(v1, M1) || M(v2, M2))
    """
    mean_diff_outer = 0.0 if is_zero(mean_diff) else torch.outer(mean_diff, mean_diff) 
    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    W1 = LW1 @ LW1.T
    tricovmat1 = trikron(U1, V1, W1)
    return torch.inverse(tricovmat1 + mean_diff_outer)

# optimal_*_diagvar
def optimal_covmat_diagvar(mean_diff, covmat1):
    """ Compute optimal diagonal variance of
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)) )
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return covmat1.diag() + mean_diff_square

def optimal_cholkroncov_diagvar(mean_diff, cholkroncov1):
    """ Compute optimal diagonal variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T)) || M(v2, diag(v3)) )
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    LU1, LV1 = cholkroncov1
    covmat1_diag = torch.kron(LU1.square().sum(1), LV1.square().sum(1))
    return covmat1_diag + mean_diff_square

def optimal_choltrikroncov_diagvar(mean_diff, choltrikroncov1):
    """ Compute optimal diagonal variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T)) || M(v2, diag(v3)) )
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    LU1, LV1, LW1 = choltrikroncov1
    covmat1_diag = trikron(LU1.square().sum(1), LV1.square().sum(1), LW1.square().sum(1))
    return covmat1_diag + mean_diff_square

# optimal_*_diagprec
def optimal_covmat_diagprec(mean_diff, covmat1):
    """ Compute optimal diagonal precision of
        argmin_v3 KL( N(v1, M1) || M(v2, diag(v3)^{-1}) )
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return 1 / (covmat1.diag() + mean_diff_square)

def optimal_cholkroncov_diagprec(mean_diff, cholkroncov1):
    """ Compute optimal diagonal variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T)) || M(v2, diag(v3)) )
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    LU1, LV1 = cholkroncov1
    covmat1_diag = torch.kron(LU1.square().sum(1), LV1.square().sum(1))
    return 1 / (covmat1_diag + mean_diff_square)

def optimal_choltrikroncov_diagprec(mean_diff, choltrikroncov1):
    """ Compute optimal diagonal variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T)) || M(v2, diag(v3)) )
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    LU1, LV1 = cholkroncov1
    covmat1_diag = trikron(LU1.square().sum(1), LV1.square().sum(1), LW1.square().sum(1))
    return 1 / (covmat1_diag + mean_diff_square)

# optimal_*_scalarvar
def optimal_covmat_scalarvar(mean_diff, covmat1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, covmat1) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return (torch.trace(covmat1) + mean_diff_square.sum()) / len(mean_diff)

def optimal_precmat_scalarvar(mean_diff, precmat1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, precmat1^{-1}) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return (torch.trace(torch.inverse(precmat1)) + mean_diff_square.sum()) / len(mean_diff)

def optimal_diagvar_scalarvar(mean_diff, diagvar1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, diagvar1) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return (torch.sum(diagvar1) + mean_diff_square.sum()) / len(mean_diff)

def optimal_diagprec_scalarvar(mean_diff, diagprec1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, diagprec1) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return (torch.sum(1/diagprec1) + mean_diff_square.sum()) / len(mean_diff)

def optimal_scalarvar_scalarvar(mean_diff, scalarvar1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, scalarvar1 I ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    d = mean_diff.size(0)
    return (d * scalarvar1 + mean_diff_square.sum()) / len(mean_diff)

def optimal_scalarprec_scalarvar(mean_diff, scalarprec1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, scalarprec1^{-1} I ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    d = mean_diff.size(0)
    return (d * scalarprec1 + mean_diff_square.sum()) / len(mean_diff)

def optimal_identity_scalarvar(mean_diff, _ignore):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, I ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    d = mean_diff.size(0)
    return (mean_diff_square.sum()) / len(mean_diff)

def optimal_cholcov_scalarvar(mean_diff, cholcov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, L1 L1^T ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    trace1 = cholcov1.square().sum()
    return (trace1 + mean_diff_square.sum()) / len(mean_diff)

def optimal_cholprec_scalarvar(mean_diff, cholcov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, L1 L1^T ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    trace1 = torch.inverse(cholcov1).square().sum()
    return (trace1 + mean_diff_square.sum()) / len(mean_diff)

def optimal_kroncov_scalarvar(mean_diff, kroncov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, (U1 kron V1) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    U1, V1 = kroncov1
    trace1 = torch.trace(U1) * torch.trace(V1)
    return (trace1 + mean_diff_square.sum()) / len(mean_diff)

def optimal_kronprec_scalarvar(mean_diff, kronprec1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, (U1 kron V1) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    U1, V1 = kronprec1
    trace1 = torch.trace(torch.inverse(U1)) * torch.trace(torch.inverse(V1))
    return (trace1 + mean_diff_square.sum()) / len(mean_diff)

def optimal_cholkroncov_scalarvar(mean_diff, cholkroncov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    LU1, LV1 = cholkroncov1
    trace1 = LU1.square().sum() * LV1.square().sum()
    return (trace1 + mean_diff_square.sum()) / len(mean_diff)

def optimal_choltrikroncov_scalarvar(mean_diff, choltrikroncov1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    LU1, LV1, LW1 = choltrikroncov1
    trace1 = LU1.square().sum() * LV1.square().sum() * LW1.square().sum()
    return (trace1 + mean_diff_square.sum()) / len(mean_diff)


def optimal_diagvarkron_scalarvar(mean_diff, diagvarkron1):
    """ Compute optimal scalar variance of
        argmin_v3 KL( N(v1, diag(a) kron diag(b) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    a, b = diagvarkron1
    trace1 = torch.outer(a, b).sum()
    return (trace1 + mean_diff_square.sum()) / len(mean_diff)

# optimal_*_scalarprec
def optimal_covmat_scalarprec(mean_diff, covmat1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, covmat1) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return len(mean_diff) / (torch.trace(covmat1) + mean_diff_square.sum()) 

def optimal_precmat_scalarprec(mean_diff, precmat1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, precmat1^{-1}) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return len(mean_diff) / (torch.trace(torch.inverse(precmat1)) + mean_diff_square.sum()) 

def optimal_diagvar_scalarprec(mean_diff, diagvar1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diagvar1) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return len(mean_diff) / (torch.sum(diagvar1) + mean_diff_square.sum()) 

def optimal_diagprec_scalarprec(mean_diff, diagprec1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diagprec1) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    return len(mean_diff) / (torch.sum(1/diagprec1) + mean_diff_square.sum()) 

def optimal_scalarprec_scalarprec(mean_diff, scalarvar1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, scalarvar1 I ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    d = mean_diff.size(0)
    return len(mean_diff) / (d * scalarvar1 + mean_diff_square.sum()) 

def optimal_scalarprec_scalarprec(mean_diff, scalarprec1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, scalarprec1^{-1} I ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    d = mean_diff.size(0)
    return len(mean_diff) / (d * scalarprec1 + mean_diff_square.sum()) 

def optimal_identity_scalarprec(mean_diff, _ignore):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, I ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    d = mean_diff.size(0)
    return len(mean_diff) / (mean_diff_square.sum()) 

def optimal_cholcov_scalarprec(mean_diff, cholcov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, L1 L1^T ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    trace1 = cholcov.square().sum(1)
    return len(mean_diff) / (trace1 + mean_diff_square.sum()) 

def optimal_cholprec_scalarprec(mean_diff, cholcov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, L1 L1^T ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    trace1 = torch.inverse(cholcov1).square().sum(1)
    return len(mean_diff) / (trace1 + mean_diff_square.sum()) 

def optimal_kroncov_scalarprec(mean_diff, kroncov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, (U1 kron V1) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    U1, V1 = kroncov1
    trace1 = torch.trace(U1) * torch.trace(V1)
    return len(mean_diff) / (trace1 + mean_diff_square.sum()) 

def optimal_kronprec_scalarprec(mean_diff, kronprec1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, (U1 kron V1) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    U1, V1 = kronprec1
    trace1 = torch.trace(torch.inverse(U1)) * torch.trace(torch.inverse(V1))
    return len(mean_diff) / (trace1 + mean_diff_square.sum()) 

def optimal_cholkroncov_scalarprec(mean_diff, cholkroncov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    LU1, LV1 = cholkroncov1
    trace1 = LU1.square().sum() * LV1.square().sum()
    return len(mean_diff) / (trace1 + mean_diff_square.sum()) 

def optimal_choltrikroncov_scalarprec(mean_diff, choltrikroncov1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    LU1, LV1, LW1 = choltrikroncov1
    trace1 = LU1.square().sum() * LV1.square().sum() * LW1.square().sum()
    return len(mean_diff) / (trace1 + mean_diff_square.sum()) 


def optimal_diagvarkron_scalarprec(mean_diff, diagvarkron1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diag(a) kron diag(b) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    a, b = diagvarkron1
    trace1 = torch.outer(a, b).sum()
    return len(mean_diff) / (trace1 + mean_diff_square.sum()) 

def optimal_diagvarrow_scalarprec(mean_diff, diagvarrow1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diag(a) kron diag(1) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    d = mean_diff.size(0)
    d1 = diagvarrow1.size(0)
    d2 = d // d1
    trace1 = diagvarrow1.sum() * d2
    return len(mean_diff) / (trace1 + mean_diff_square.sum())

def optimal_diagvarcol_scalarprec(mean_diff, diagvarcol1):
    """ Compute optimal scalar precision of
        argmin_v3 KL( N(v1, diag(1) kron diag(b) ) || M(v2, v3 I))
    """
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    d = mean_diff.size(0)
    d2 = diagvarcol1.size(0)
    d1 = d // d2
    trace1 = d1 * diagvarcol1.sum()
    return len(mean_diff) / (trace1 + mean_diff_square.sum()) 

# optimal_*_diagvarrow
def optimal_covmat_diagvarrow(mean_diff, covmat1, dims):
    """ Compute optimal row diagonal of
        argmin_v3 KL( N(v1, M1 ) || M(v2, diag(v3) kron I))
    """
    d1, d2 = dims
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()

    mean_diff_rows = mean_diff_square.view(d1, d2).sum(1)
    covmat1_rows = torch.diagonal(covmat1.view(d1, d2, d1, d2), dim1=1, dim2=3).sum(-1)

    return (covmat1_rows.diag() + mean_diff_rows) / d2

def optimal_cholkroncov_diagvarrow(mean_diff, cholkroncov1, dims):
    """ Compute optimal row diagonal of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T) ) || M(v2, diag(v3) kron I))
    """
    d1, d2 = dims
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    mean_diff_rows = mean_diff_square.view(d1, d2).sum(1)

    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T

    covmat1_rows = U1 * LV1.square().sum()

    return (covmat1_rows.diag() + mean_diff_rows) / d2

def optimal_choltrikroncov_diagvarrow(mean_diff, choltrikroncov1, dims):
    """ Compute optimal row diagonal of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T) ) || M(v2, diag(v3) kron I))
    """
    d1, d2, d3 = dims
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    mean_diff_rows = mean_diff_square.view(d1, d2, d3).sum((1, 2))

    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T

    covmat1_rows = U1 * LV1.square().sum() * LW1.square().sum()

    return (covmat1_rows.diag() + mean_diff_rows) / d2 / d3

# optimal_*_diagvarcol
def optimal_covmat_diagvarcol(mean_diff, covmat1, dims):
    """ Compute optimal row diagonal of
        argmin_v3 KL( N(v1, M1 ) || M(v2, I kron diag(v3)))
    """
    d1, d2 = dims
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    mean_diff_cols = mean_diff_square.view(d1, d2).sum(0)
    covmat1_cols = torch.diagonal(covmat1.view(d1, d2, d1, d2), dim1=0, dim2=2).sum(-1)

    return (covmat1_cols.diag() + mean_diff_cols) / d1

def optimal_cholkroncov_diagvarcol(mean_diff, cholkroncov1, dims):
    """ Compute optimal column diagonal of
        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T) ) || M(v2, I kron diag(v3)))
    """
    d1, d2 = dims
    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
    mean_diff_cols = mean_diff_square.view(d1, d2).sum(0)

    LU1, LV1 = cholkroncov1
    V1 = LV1 @ LV1.T

    covmat1_cols = V1 * LU1.square().sum()

    return (covmat1_cols.diag() + mean_diff_cols) / d1

# TODO: this is tricky because quadratic_diagvarcol is broken for tensors at the moment
# Fixing this requires rethinking how quadratic_* has access to all dimensions to make this work.
# One option would be to not vectorize the means, but this change would significantly alter the outer interface.
#def optimal_choltrikroncov_diagvarcol(mean_diff, choltrikroncov1, dims):
#    """ Compute optimal column diagonal of
#        argmin_v3 KL( N(v1, (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T) ) || M(v2, I kron diag(v3)))
#    """
#    d1, d2, d3 = dims
#    mean_diff_square = 0.0 if is_zero(mean_diff) else mean_diff.square()
#    mean_diff_cols = mean_diff_square.view(d1, d2, d3).sum((0, 2))
#
#    LU1, LV1, LW1 = choltrikroncov1
#    V1 = LV1 @ LV1.T
#
#    covmat1_cols = V1 * LU1.square().sum() * LW1.square().sum()
#
#    return (covmat1_cols.diag() + mean_diff_cols) / d1 / d3


def optimal_covariance(mean1, cov_type1, cov1, mean2, cov_type2, dims=None):
    """ Find optimal cov2 that solves
            argmin_cov2 KL( N(mean1, cov1) || N(mean2, cov2))
    """
    func_name = f"optimal_{cov_type1}_{cov_type2}"

    func = globals()[func_name]

    mean_diff = mean1 - mean2

    if cov_type2 in ['diagvarkron', 'diagvarrow', 'diagvarcol']:
        assert dims is not None, f"optimal_covariance: {cov_type2} requires dimensions when computing optimal variance (Got: d1={d1}, d2={d2})"
        return func(mean_diff, cov1, dims)
    else:
        return func(mean_diff, cov1)





import torch
import math

import normalkl.trace
import normalkl.logdet
import normalkl.quadratic

from normalkl.util import dim_dtype_device_of_cov, is_zero


def trace_term_func(cov_type1, cov1, cov_type2, cov2):
    func_name = f"trace_{cov_type1}_{cov_type2}"
    trace_func = getattr(normalkl.trace, func_name)
    
    if trace_func:
        return trace_func(cov1, cov2)
    else:
        raise NotImplementedError(f"trace term not implemented yet for covariances of type '{cov_type1}' to '{cov_type2}' (please add a pull request on Github)")
    

def volume_term_func(cov_type1, cov1, cov_type2, cov2, total_dim=None):
    q_logdet_func_name = f"logdet_{cov_type1}"
    p_logdet_func_name = f"logdet_{cov_type2}"

    q_logdet_func = getattr(normalkl.logdet, q_logdet_func_name, None)
    p_logdet_func = getattr(normalkl.logdet, p_logdet_func_name, None)
    
    if q_logdet_func:
        if any([substring in q_logdet_func_name for substring in ['scalar', 'covrow', 'covcol']]):
            q_logdet = q_logdet_func(cov1, total_dim)
        else:
            q_logdet = q_logdet_func(cov1)
    else:
        raise NotImplementedError(f"log determinant not implemented yet for covariances of type '{cov_type1}' (please add a pull request on Github)")
    
    if p_logdet_func:
        if any([substring in p_logdet_func_name for substring in ['scalar', 'row', 'col']]):
            p_logdet = p_logdet_func(cov2, total_dim)
        else:
            p_logdet = p_logdet_func(cov2)
    else:
        raise NotImplementedError(f"log determinant not implemented yet for covariances of type '{cov_type2}' (please add a pull request on Github)")

    return 0.5 * (p_logdet - q_logdet)


def quadratic_term_func(mean_diff, cov_type2, cov2):
    func_name = f"quadratic_{cov_type2}"

    quadratic_func = getattr(normalkl.quadratic, func_name)

    if quadratic_func:
        return quadratic_func(mean_diff, cov2)
    else:
        raise NotImplementedError(f"quadratic term not implemented yet for covariances of type '{cov_type2}' (please add a pull request on Github)")
    

def kl(mean1, cov_type1, cov1, mean2, cov_type2, cov2, verbose=False):
    """ Computes the KL divergence between two multivariate normals
            KL( N(mean1, cov1) || N(mean2, cov2) )

        Input:
            mean1: mean of first multivariate normal 
            cov_typ1: covariance type of first multivariate normal
            cov1: covariance of first multivariate normal
            mean2: mean of first multivariate normal 
            cov_typ2: covariance type of first multivariate normal
            cov2: covariance of first multivariate normal

        Returns: the KL (scalar number)
    """ 

    total_dim, dtype, device = dim_dtype_device_of_cov(mean1, cov_type1, cov1)

    if isinstance(mean1, float):
        mean1 = torch.zeros(total_dim, dtype=dtype, device=device)
    if isinstance(mean2, float):
        mean2 = torch.zeros(total_dim, dtype=dtype, device=device)

    trace_term = trace_term_func(cov_type1, cov1, cov_type2, cov2)
    volume_term = volume_term_func(cov_type1, cov1, cov_type2, cov2, total_dim=total_dim)
    quadratic_term = quadratic_term_func(mean1 - mean2, cov_type2, cov2)
    dimension_term = -0.5 * total_dim

    kl_div = trace_term + volume_term + quadratic_term + dimension_term
    kl_div = kl_div.squeeze()

    if verbose:
        print(f"KL = {kl_div.item():.2f}  [trace: {trace_term.item():.2f}, volume: {volume_term.item():.2f}, quadratic: {quadratic_term.item()}, dimension: {dimension_term:.2f}")

    return kl_div




import torch

def dim_dtype_device_of_cov(mean, cov_type, cov):
    if cov_type in ['covmat', 'precmat', 'diagvar', 'diagprec', 'cholcov', 'cholprec']:
        return cov.size(0), cov.dtype, cov.device
    elif cov_type in ['scalarvar', 'scalarprec', 'identity', 'diagvarrow', 'diagvarcol']:
        if type(mean) == float:
            raise ValueError(f"Type '{cov_type}' requires mean tensor and not just float, because dimensionality can not be known in this case.")
        return mean.size(0), mean.dtype, mean.device
    elif cov_type in ['kroncov', 'kronprec', 'cholkroncov', 'diagvarkron']:
        return cov[0].size(0) * cov[1].size(0), cov[0].dtype, cov[0].device
    elif cov_type in ['choltrikroncov']:
        return cov[0].size(0) * cov[1].size(0) * cov[2].size(0), cov[0].dtype, cov[0].device
    else:
        raise NotImplementedError(f"Unknown cov_type: {cov_type}")


def is_zero(x):
    if isinstance(x, float):
        return x == 0.0
    elif isinstance(x, torch.Tensor):
        return torch.all(x == 0.0)
    else:
        return False



import unittest

import torch

from normalkl import kl, auto_kl
from normalkl import optimal_covariance

trikron = lambda A, B, C: torch.kron(A, torch.kron(B, C))

class TestAutoKL(unittest.TestCase):
    def setUp(self):
        dtype = torch.float64

        d1, d2 = 3, 4
        d = d1 * d2

        self.d1 = d1
        self.d2 = d2
        self.d = d

        tensor_d1, tensor_d2, tensor_d3 = 2, 2, 3
        self.tensor_d1 = tensor_d1
        self.tensor_d2 = tensor_d2
        self.tensor_d3 = tensor_d3

        dtype=torch.float64

        # Means
        self.mean1 = torch.randn(d, dtype=dtype)
        self.mean2 = torch.randn(d, dtype=dtype)

        # Full covariance 
        self.L1 = torch.randn(d, d, dtype=dtype).tril()
        self.L2 = torch.randn(d, d, dtype=dtype).tril()
        self.covmat1 = self.L1 @ self.L1.T
        self.precmat1 = torch.inverse(self.covmat1)
        self.covmat2 = self.L2 @ self.L2.T
        self.precmat2 = torch.inverse(self.covmat2)
            
        # Cholesky factors for cholkroncov1 and cholkroncov2
        self.LU1 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV1 = torch.randn(d2, d2, dtype=dtype).tril()
        self.LU2 = torch.randn(d1, d1, dtype=dtype).tril()
        self.LV2 = torch.randn(d2, d2, dtype=dtype).tril()

        self.tensor_LU1 = torch.randn(tensor_d1, tensor_d1, dtype=dtype).tril()
        self.tensor_LV1 = torch.randn(tensor_d2, tensor_d2, dtype=dtype).tril()
        self.tensor_LW1 = torch.randn(tensor_d3, tensor_d3, dtype=dtype).tril()
        self.tensor_LU2 = torch.randn(tensor_d1, tensor_d1, dtype=dtype).tril()
        self.tensor_LV2 = torch.randn(tensor_d2, tensor_d2, dtype=dtype).tril()
        self.tensor_LW2 = torch.randn(tensor_d3, tensor_d3, dtype=dtype).tril()

        # Covariance matrices for cholkroncov1 and cholkroncov2
        self.U1 = self.LU1 @ self.LU1.T
        self.V1 = self.LV1 @ self.LV1.T
        self.U2 = self.LU2 @ self.LU2.T
        self.V2 = self.LV2 @ self.LV2.T
        self.tensor_U1 = self.tensor_LU1 @ self.tensor_LU1.T
        self.tensor_V1 = self.tensor_LV1 @ self.tensor_LV1.T
        self.tensor_W1 = self.tensor_LW1 @ self.tensor_LW1.T
        self.tensor_U2 = self.tensor_LU2 @ self.tensor_LU2.T
        self.tensor_V2 = self.tensor_LV2 @ self.tensor_LV2.T
        self.tensor_W2 = self.tensor_LW2 @ self.tensor_LW2.T

        # Cholesky Kronecker decompositions
        self.cholkroncov1 = (self.LU1, self.LV1)
        self.cholkroncov2 = (self.LU2, self.LV2)
        self.choltrikroncov1 = (self.tensor_LU1, self.tensor_LV1, self.tensor_LW1)
        self.choltrikroncov2 = (self.tensor_LU2, self.tensor_LV2, self.tensor_LW1)

        # Full covariance matrices
        self.kron_covmat1 = torch.kron(self.U1, self.V1)
        self.kron_covmat2 = torch.kron(self.U2, self.V2)
        self.trikron_covmat1 = trikron(self.tensor_U1, self.tensor_V1, self.tensor_W1)
        self.trikron_covmat2 = trikron(self.tensor_U2, self.tensor_V2, self.tensor_W2)

        # Other inputs
        self.I = torch.eye(d, dtype=dtype)
        self.v1 = torch.randn(d, dtype=dtype)
        self.v2 = torch.randn(d, dtype=dtype)
        self.s2 = torch.tensor(2.0, dtype=dtype)
        self.a1 = torch.randn(d1, dtype=dtype).square()
        self.b1 = torch.randn(d2, dtype=dtype).square()

        self.rtol = 1e-6
        self.atol = 1e-6


    # auto_kl_*_covmat
    def test_auto_kl_covmat_covmat(self):
        optimal_covmat2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'covmat')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'covmat', optimal_covmat2, verbose=True)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'covmat')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_precmat_covmat(self):
        optimal_covmat2 = optimal_covariance(self.mean1, 'precmat', self.precmat1, self.mean2, 'covmat')
        result_manual = kl(self.mean1, 'precmat', self.precmat1, self.mean2, 'covmat', optimal_covmat2, verbose=True)
        result_auto = auto_kl(self.mean1, 'precmat', self.precmat1, self.mean2, 'covmat')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_covmat(self):
        optimal_covmat2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'covmat')
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'covmat', optimal_covmat2, verbose=True)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'covmat')

        self.assertTrue(torch.allclose(result_manual, result_auto))


    # auto_kl_*_precmat
    def test_auto_kl_covmat_precmat(self):
        optimal_precmat2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'precmat')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'precmat', optimal_precmat2, verbose=True)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'precmat')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_precmat_precmat(self):
        optimal_precmat2 = optimal_covariance(self.mean1, 'precmat', self.precmat1, self.mean2, 'precmat')
        result_manual = kl(self.mean1, 'precmat', self.precmat1, self.mean2, 'precmat', optimal_precmat2, verbose=True)
        result_auto = auto_kl(self.mean1, 'precmat', self.precmat1, self.mean2, 'precmat')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_precmat(self):
        optimal_precmat2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'precmat')
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'precmat', optimal_precmat2, verbose=True)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'precmat')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_diagvar
    def test_auto_kl_covmat_diagvar(self):
        optimal_diagvar2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar', optimal_diagvar2, verbose=True)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_diagvar(self):
        optimal_diagvar2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvar')
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvar', optimal_diagvar2, verbose=True)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_diagvar_zero(self):
        optimal_diagvar2 = optimal_covariance(0.0, 'cholkroncov', self.cholkroncov1, 0.0, 'diagvar')
        result_manual = kl(0.0, 'cholkroncov', self.cholkroncov1, 0.0, 'diagvar', optimal_diagvar2, verbose=True)
        result_auto = auto_kl(0.0, 'cholkroncov', self.cholkroncov1, 0.0, 'diagvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_choltrikroncov_diagvar(self):
        optimal_diagvar2 = optimal_covariance(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'diagvar')
        result_manual = kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'diagvar', optimal_diagvar2, verbose=True)
        result_auto = auto_kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'diagvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_diagprec
    def test_auto_kl_covmat_diagprec(self):
        optimal_diagprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec', optimal_diagprec2)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagprec')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_scalarvar
    def test_auto_kl_covmat_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar', optimal_scalarvar2)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholcov_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'cholcov', self.L1, self.mean2, 'scalarvar')
        result_manual = kl(self.mean1, 'cholcov', self.L1, self.mean2, 'scalarvar', optimal_scalarvar2)
        result_auto = auto_kl(self.mean1, 'cholcov', self.L1, self.mean2, 'scalarvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))


    def test_auto_kl_cholkroncov_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar')
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar', optimal_scalarvar2)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_choltrikroncov_scalarvar(self):
        optimal_scalarvar2 = optimal_covariance(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'scalarvar')
        result_manual = kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'scalarvar', optimal_scalarvar2)
        result_auto = auto_kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'scalarvar')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_scalarprec
    def test_auto_kl_covmat_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec')
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec', optimal_scalarprec2)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'scalarprec')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec')
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec', optimal_scalarprec2)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'scalarprec')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_scalarprec(self):
        optimal_scalarprec2 = optimal_covariance(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'scalarprec')
        result_manual = kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'scalarprec', optimal_scalarprec2)
        result_auto = auto_kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'scalarprec')

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_diagvarrow
    def test_auto_kl_covmat_diagvarrow(self):
        optimal_diagvarrow2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarrow', dims=(self.d1, self.d2))
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarrow', optimal_diagvarrow2)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarrow', dims=(self.d1, self.d2))

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_cholkroncov_diagvarrow(self):
        optimal_diagvarrow2 = optimal_covariance(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvarrow', dims=(self.d1, self.d2))
        result_manual = kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvarrow', optimal_diagvarrow2)
        result_auto = auto_kl(self.mean1, 'cholkroncov', self.cholkroncov1, self.mean2, 'diagvarrow', dims=(self.d1, self.d2))

        self.assertTrue(torch.allclose(result_manual, result_auto))

    def test_auto_kl_choltrikroncov_diagvarrow(self):
        optimal_diagvarrow2 = optimal_covariance(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'diagvarrow', dims=(self.tensor_d1, self.tensor_d2, self.tensor_d3))
        result_manual = kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'diagvarrow', optimal_diagvarrow2)
        result_auto = auto_kl(self.mean1, 'choltrikroncov', self.choltrikroncov1, self.mean2, 'diagvarrow', dims=(self.tensor_d1, self.tensor_d2, self.tensor_d3))

        self.assertTrue(torch.allclose(result_manual, result_auto))

    # auto_kl_*_diagvarcol
    def test_auto_kl_covmat_diagvarcol(self):
        optimal_diagvarrow2 = optimal_covariance(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarrow', dims=(self.d1, self.d2))
        result_manual = kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarrow', optimal_diagvarrow2)
        result_auto = auto_kl(self.mean1, 'covmat', self.covmat1, self.mean2, 'diagvarrow', dims=(self.d1, self.d2))

        self.assertTrue(torch.allclose(result_manual, result_auto))



# Run the tests
if __name__ == '__main__':
    unittest.main(argv=[''], exit=False)




from setuptools import setup, find_packages

setup(
    name='normalkl',         # name of the package
    version='0.1.15',                     # version number
    description='A PyTorch package for computing KL divergences between normal distributions.',  # short description
    long_description=open('README.md').read(),  # detailed description (from README)
    long_description_content_type='text/markdown',
    author='Tycho van der Ouderaa',
    author_email='tychovdo@gmail.com',
    url='https://github.com/tychovdo/normal-kl',  # project URL
    license='MIT',
    packages=find_packages(),          # automatically finds your package
    install_requires=[
        'torch',                       # specify PyTorch as a dependency
        # other dependencies
    ],
    classifiers=[                      # optional classifiers
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',           # specify Python version compatibility
)



""" Functions that compute trace term of the KL divergence for different covariance types. """

import torch

def trikron(A, B, C):
    return torch.kron(A, torch.kron(B, C))

# trace_covmat_*

def trace_covmat_covmat(covmat1, covmat2):
    """ Computes trace term: Tr(M2^{-1} M1) / 2 """
    return 0.5 * torch.trace(torch.inverse(covmat2) @ covmat1)
    
def trace_covmat_precmat(covmat1, precmat2):
    """ Computes trace term: Tr(M2 M1) / 2 """
    return 0.5 * torch.trace(precmat2 @ covmat1)
    
def trace_covmat_diagvar(covmat1, diagvar2):
    """ Computes trace term: Tr(diag(v2)^{-1} M1) / 2 """
    return 0.5 * torch.sum(covmat1.diag() / diagvar2)

def trace_covmat_diagprec(covmat1, diagprec2):
    """ Computes trace term: Tr(diag(v2) M1) / 2 """
    return 0.5 * torch.sum(covmat1.diag() * diagprec2)

def trace_covmat_scalarvar(covmat1, scalarvar2):
    """ Computes trace term: Tr(scalarvar2^{-1} covmat1) / 2 """
    return 0.5 * torch.trace(covmat1) / scalarvar2.abs().squeeze()

def trace_covmat_scalarprec(covmat1, scalarprec2):
    """ Computes trace term: Tr(scalarprec2 I covmat1) / 2 """
    return 0.5 * torch.trace(covmat1) * scalarprec2.abs().squeeze()

def trace_covmat_identity(covmat1, _):
    """ Computes trace term: Tr(I covmat1)/2 """
    return 0.5 * torch.trace(covmat1)

def trace_covmat_cholcov(covmat1, cholcov2):
    """ Computes trace term: Tr((L2 L2^T)^{-1} M1) / 2 """
    L2_inv = torch.inverse(cholcov2)
    return 0.5 * torch.trace(L2_inv.T @ L2_inv @ covmat1)
    
def trace_covmat_cholprec(covmat1, cholprec2):
    """ Computes trace term: Tr((L2 L2^T) M1) / 2 """
    return 0.5 * torch.trace(cholprec2 @ cholprec2.T @ covmat1)

def trace_covmat_kroncov(covmat1, kroncov2):
    """ Computes trace term: Tr((U2 kron V2)^{-1} M1) / 2 """
    U2, V2 = kroncov2
    UV2 = torch.kron(U2, V2)
    return 0.5 * torch.trace(torch.linalg.solve(UV2, covmat1))
    
def trace_covmat_kronprec(covmat1, kronprec2):
    """ Computes trace term: Tr((U2 kron V2) M1) / 2 """
    U2, V2 = kronprec2
    UV2 = torch.kron(U2, V2)
    return 0.5 * torch.trace(UV2 @ covmat1)

def trace_covmat_cholkroncov(covmat1, cholkroncov2):
    """ Computes trace term: Tr((LU2 LU2^T kron LV2 LV2^T)^{-1} M1) / 2 """
    LU2, LV2 = cholkroncov2
    U2 = LU2 @ LU2.T
    V2 = LV2 @ LV2.T
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    
    return 0.5 * torch.trace(torch.kron(U2_inv, V2_inv) @ covmat1)

def trace_covmat_choltrikroncov(covmat1, choltrikroncov2):
    """ Computes trace term: Tr((LU2 LU2^T kron LV2 LV2^T kron LW2 LW2^T)^{-1} M1) / 2 """
    LU2, LV2, LW2 = choltrikroncov2
    U2 = LU2 @ LU2.T
    V2 = LV2 @ LV2.T
    W2 = LW2 @ LW2.T
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    W2_inv = torch.inverse(W2)
    
    return 0.5 * torch.trace(trikron(U2_inv, V2_inv, W2_inv) @ covmat1)
    
def trace_covmat_diagvarkron(covmat1, diagvarkron2):
    """ Computes Tr(diag(vec(ba^T))^{-1} M1)/2 """
    diagvarrow, diagvarcol = diagvarkron2
    kron = torch.outer(diagvarrow, diagvarcol).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / kron)

def trace_covmat_diagvarrow(covmat1, diagvarrow2):
    """ Computes Tr(diag(vec(1a^T))^{-1} M1)/2 """
    d1 = len(diagvarrow2)
    d2 = covmat1.size(0) // d1
    kron = diagvarrow2.view(-1, 1).repeat(1, d2).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / kron)

def trace_covmat_diagvarcol(covmat1, diagvarcol2):
    """ Computes Tr(diag(vec(b1^T))^{-1} M1)/2 """
    d2 = len(diagvarcol2)
    d1 = covmat1.size(0) // d2
    kron = diagvarcol2.view(1, -1).repeat(d1, 1).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / kron)

# trace_precmat_*

def trace_precmat_covmat(precmat1, covmat2):
    """ Computes trace term: Tr(M2^{-1} M1) / 2 """
    return 0.5 * torch.trace(torch.inverse(covmat2) @ torch.inverse(precmat1))
    
def trace_precmat_precmat(precmat1, precmat2):
    """ Computes trace term: Tr(M2 M1) / 2 """
    return 0.5 * torch.trace(precmat2 @ torch.inverse(precmat1))

# trace_cholcov_*

def trace_cholcov_scalarvar(cholcov1, scalarvar2):
    """ Computes trace term: Tr(scalarvar2^{-1} cholcov1) / 2 """
    return 0.5 * torch.sum(cholcov1 ** 2) / scalarvar2.abs().squeeze()

def trace_cholcov_cholcov(cholcov1, cholcov2):
    """ Computes trace term: Tr((L2 L2^T)^{-1} (L1 L1^T)) / 2 """

    L2_inv = torch.inverse(cholcov2)
    C = cholcov1.T @ L2_inv.T
    return 0.5 * (C ** 2).sum()

# trace_cholkroncov_*

def trace_cholkroncov_covmat(cholkroncov1, covmat2):
    """ Computes trace term: Tr(M2^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    return 0.5 * torch.trace(torch.inverse(covmat2) @ torch.kron(U1, V1))

def trace_cholkroncov_precmat(cholkroncov1, precmat2):
    """ Computes trace term: Tr(M2 (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    return 0.5 * torch.trace(precmat2 @ torch.kron(U1, V1))

def trace_cholkroncov_diagvar(cholkroncov1, diagvar2):
    """ Computes trace term: Tr(diag(v2)^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U1_diag = LU1.square().sum(1)
    V1_diag = LV1.square().sum(1)
    return 0.5 * torch.sum(torch.outer(U1_diag, V1_diag).view(-1) / diagvar2)

def trace_cholkroncov_diagprec(cholkroncov1, diagprec2):
    """ Computes trace term: Tr(diag(v2) (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U1_diag = LU1.square().sum(1)
    V1_diag = LV1.square().sum(1)
    return 0.5 * torch.sum(torch.outer(U1_diag, V1_diag).view(-1)  * diagprec2)

def trace_cholkroncov_scalarvar(cholkroncov1, scalarvar2):
    """ Computes trace term: Tr(scalarvar2^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    return 0.5 * torch.sum(LU1 ** 2) * torch.sum(LV1 ** 2) / scalarvar2

def trace_cholkroncov_scalarprec(cholkroncov1, scalarprec2):
    """ Computes trace term: Tr(scalarprec2 I (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    return 0.5 * torch.sum(LU1 ** 2) * torch.sum(LV1 ** 2) * scalarprec2

def trace_cholkroncov_identity(cholkroncov1, _):
    """ Computes trace term: Tr(I (LU1 LU1^T kron LV1 LV1^T))/2 """
    LU1, LV1 = cholkroncov1
    return 0.5 * torch.trace(LU1 @ LU1.T) * torch.trace(LV1 @ LV1.T)

def trace_cholkroncov_cholcov(cholkroncov1, cholcov2):
    """ Computes trace term: Tr((L2 L2.T)^{-1} (LU1 LU1.T kron LV1 LV1.T )) / 2 """
    L2_inv = torch.inverse(cholcov2)
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    return 0.5 * torch.trace(L2_inv.T @ L2_inv @ torch.kron(U1, V1))

def trace_cholkroncov_cholprec(cholkroncov1, cholprec2):
    """ Computes trace term: Tr((L2 L2.T) (LU1 LU1.T kron LV1 LV1.T )) / 2 """
    L2 = cholprec2
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    return 0.5 * torch.trace(L2 @ L2.T @ torch.kron(U1, V1))

def trace_cholkroncov_kroncov(cholkroncov1, kroncov2):
    """ Computes trace term: Tr((U2 kron V2)^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U2, V2 = kroncov2
    U2_inv = torch.inverse(U2)
    V2_inv = torch.inverse(V2)
    return 0.5 * torch.trace(LU1.T @ U2_inv @ LU1) * torch.trace(LV1.T @ V2_inv @ LV1)

def trace_cholkroncov_kronprec(cholkroncov1, kronprec2):
    """ Computes trace term: Tr((U2 kron V2) (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    U2, V2 = kronprec2
    return 0.5 * torch.trace(LU1.T @ U2 @ LU1) * torch.trace(LV1.T @ V2 @ LV1)

def trace_cholkroncov_cholkroncov(cholkroncov1, cholkroncov2):
    """ Computes trace term: Tr((LU2 LU2^T kron LV2 LV2^T)^{-1} (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    LU2, LV2 = cholkroncov2
    LU2_inv = torch.inverse(LU2)
    LV2_inv = torch.inverse(LV2)

    LU1TLU2_invT = LU1.T @ LU2_inv.T
    trace_LU1TLU2_invT = LU1TLU2_invT.square().sum()
    LV1TLV2_invT = LV1.T @ LV2_inv.T
    trace_LV1TLV2_invT = LV1TLV2_invT.square().sum()

    return 0.5 * trace_LU1TLU2_invT * trace_LV1TLV2_invT

def trace_cholkroncov_cholkronprec(cholkroncov1, cholkronprec2):
    """ Computes trace term: Tr((LU2 LU2^T kron LV2 LV2^T) (LU1 LU1^T kron LV1 LV1^T)) / 2 """
    LU1, LV1 = cholkroncov1
    LU2, LV2 = cholkronprec2
    LU1TLU2 = LU1.T @ LU2
    trace_U2U1 = LU1TLU2.square().sum()
    LV1TLV2 = LV1.T @ LV2
    trace_V2V1 = LV1TLV2.square().sum()

    return 0.5 * trace_U2U1 * trace_V2V1

def trace_cholkroncov_diagvarrow(cholkroncov1, diagvarrow2): # TODO: avoid kronecker here, can be much faster
    """ Computes Tr(diag(vec(1a^T))^{-1} (LU1 LU1^T kron LV1 LV1^T))/2 """
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    covmat1 = torch.kron(U1, V1)
    d1 = len(U1)
    d2 = len(V1)

    kron = diagvarrow2.view(-1, 1).repeat(1, d2).view(-1)
    print('d', d1, d2, diagvarrow2.shape)
    print(covmat1.shape, kron.shape, 444)
    return 0.5 * torch.sum(covmat1.diag() / kron)

def trace_cholkroncov_diagvarcol(cholkroncov1, diagvarcol2): # TODO: avoid kronecker here, can be much faster
    """ Computes Tr(diag(vec(b1^T))^{-1} (LU1LU1^T kron LV1 LV1^T))/2 """
    LU1, LV1 = cholkroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    covmat1 = torch.kron(U1, V1)
    d1 = len(U1)
    d2 = len(V1)

    kron = diagvarcol2.view(1, -1).repeat(d1, 1).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / kron)

# trace_choltrikroncov_*

def trace_choltrikroncov_covmat(choltrikroncov1, covmat2):
    """ Computes trace term: Tr(M2^{-1} (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T)) / 2 """
    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    W1 = LW1 @ LW1.T
    return 0.5 * torch.trace(torch.inverse(covmat2) @ trikron(U1, V1, W1))

def trace_choltrikroncov_precmat(choltrikroncov1, precmat2):
    """ Computes trace term: Tr(M2 (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T)) / 2 """
    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    W1 = LW1 @ LW1.T
    return 0.5 * torch.trace(precmat2 @ trikron(U1, V1, W1))

def trace_choltrikroncov_diagvar(choltrikroncov1, diagvar2):
    """ Computes trace term: Tr(diag(v2)^{-1} (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T)) / 2 """
    LU1, LV1, LW1 = choltrikroncov1
    U1_diag = LU1.square().sum(1)
    V1_diag = LV1.square().sum(1)
    W1_diag = LW1.square().sum(1)
    return 0.5 * torch.sum(torch.einsum('a,b,c->abc',U1_diag, V1_diag, W1_diag).view(-1) / diagvar2)

def trace_choltrikroncov_diagprec(choltrikroncov1, diagprec2):
    """ Computes trace term: Tr(diag(v2) (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T)) / 2 """
    LU1, LV1, LW1 = choltrikroncov1
    U1_diag = LU1.square().sum(1)
    V1_diag = LV1.square().sum(1)
    W1_diag = LW1.square().sum(1)
    return 0.5 * torch.sum(torch.einsum('a,b,c->abc',U1_diag, V1_diag, W1_diag).view(-1) * diagprec2)

def trace_choltrikroncov_scalarvar(choltrikroncov1, scalarvar2):
    """ Computes trace term: Tr(scalarvar2^{-1} (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T)) / 2 """
    LU1, LV1, LW1 = choltrikroncov1
    return 0.5 * torch.sum(LU1 ** 2) * torch.sum(LV1 ** 2) * torch.sum(LW1 ** 2) / scalarvar2

def trace_choltrikroncov_scalarprec(choltrikroncov1, scalarprec2):
    """ Computes trace term: Tr(scalarprec2 I (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T)) / 2 """
    LU1, LV1, LW1 = choltrikroncov1
    return 0.5 * torch.sum(LU1 ** 2) * torch.sum(LV1 ** 2) * torch.sum(LW1 ** 2) * scalarprec2

def trace_choltrikroncov_identity(choltrikroncov1, _):
    """ Computes trace term: Tr(I (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T))/2 """
    LU1, LV1, LW1 = choltrikroncov1

    return 0.5 * torch.sum(LU1 ** 2) * torch.sum(LV1 ** 2) * torch.sum(LW1 ** 2)

def trace_choltrikroncov_cholcov(choltrikroncov1, cholcov2):
    """ Computes trace term: Tr((L2 L2.T)^{-1} (LU1 LU1.T kron LV1 LV1.T kron LW1 LW1.T)) / 2 """
    L2_inv = torch.inverse(cholcov2)
    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    W1 = LW1 @ LW1.T
    return 0.5 * torch.trace(L2_inv.T @ L2_inv @ trikron(U1, V1, W1))

def trace_choltrikroncov_cholprec(choltrikroncov1, cholprec2):
    """ Computes trace term: Tr((L2 L2.T) (LU1 LU1.T kron LV1 LV1.T )) / 2 """
    L2 = cholprec2
    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    W1 = LW1 @ LW1.T
    return 0.5 * torch.trace(L2 @ L2.T @ trikron(U1, V1, W1))

def trace_choltrikroncov_diagvarrow(choltrikroncov1, diagvarrow2): # TODO: avoid kronecker here, can be much faster
    """ Computes Tr(diag(vec(1a^T))^{-1} (LU1 LU1^T kron LV1 LV1^T kron LW1 LW1^T))/2 """
    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    W1 = LW1 @ LW1.T
    covmat1 = trikron(U1, V1, W1)
    d1 = len(U1)
    d2 = len(V1)
    d3 = len(W1)

    kron = diagvarrow2.view(-1, 1, 1).repeat(1, d2, d3).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / kron)

def trace_choltrikroncov_diagvarcol(choltrikroncov1, diagvarcol2): # TODO: avoid kronecker here, can be much faster
    """ Computes Tr(diag(vec(b1^T))^{-1} (LU1LU1^T kron LV1 LV1^T kron LW1 LW1^T))/2 """
    LU1, LV1, LW1 = choltrikroncov1
    U1 = LU1 @ LU1.T
    V1 = LV1 @ LV1.T
    W1 = LW1 @ LW1.T
    covmat1 = trikron(U1, V1, W1)
    d1 = len(U1)
    d2 = len(V1)
    d3 = len(W1)

    kron = diagvarcol2.view(1, -1, 1).repeat(d1, 1, d3).view(-1)
    return 0.5 * torch.sum(covmat1.diag() / kron)


